/*
 * Copyright (C) 2012 CERN (www.cern.ch)
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/fmc.h>
#include <linux/interrupt.h>
#include <linux/moduleparam.h>
#include <linux/gpio.h>
#include <linux/fmc-sdb.h>
#include "spec.h"

/* Name for FMC DIO */
#define DIO_FMC_ID_NAME "FmcDio5cha"

/* The main role of this file is offering the fmc_operations for the spec */

static int spec_validate(struct fmc_device *fmc, struct fmc_driver *drv)
{
	struct spec_dev *spec = fmc->carrier_data;
	struct pci_dev *pdev = spec->pdev;
	int busid = (pdev->bus->number << 8) | pdev->devfn;
	int i;

	if (!drv->busid_n)
		return 0; /* everyhing is valid */
	for (i = 0; i < drv->busid_n; i++)
		if (drv->busid_val[i] == busid)
			return i;
	return -ENOENT;
}

static int spec_reprogram_raw(struct fmc_device *fmc, struct fmc_driver *drv,
			      void *gw, unsigned long len)
{
	struct spec_dev *spec = fmc->carrier_data;
	struct device *dev = fmc->hwdev;
	int ret;

	if (!gw || !len) {
		dev_err(dev, "Invalid firmware buffer - buf: %p len: %ld\n",
			gw, len);
		return -EINVAL;
	}

	if (!drv)
		dev_info(dev, "Carrier FPGA re-program\n");

	fmc_free_sdb_tree(fmc);

	ret = spec_load_fpga(spec, gw, len);
	if (ret < 0) {
		dev_err(dev, "writing firmware: error %i\n", ret);
		return ret;
	}

	return 0;
}

static int spec_reprogram(struct fmc_device *fmc, struct fmc_driver *drv,
			  char *gw)
{
	const struct firmware *fw;
	struct device *dev = fmc->hwdev;
	int ret;

	if (!gw)
		gw = spec_fw_name;

	if (!strlen(gw)) { /* use module parameters from the driver */
		int index;

		index = spec_validate(fmc, drv);

		gw = drv->gw_val[index];
		if (!gw)
			return -ESRCH; /* the caller may accept this */
	}

	dev_info(fmc->hwdev, "reprogramming with %s\n", gw);
	ret = request_firmware(&fw, gw, dev);
	if (ret < 0) {
		dev_warn(dev, "request firmware \"%s\": error %i\n", gw, ret);
		goto out;
	}

	ret = spec_reprogram_raw(fmc, drv, (void *)fw->data, fw->size);
	if (ret < 0) {
		dev_err(dev, "write firmware \"%s\": error %i\n", gw, ret);
		goto out;
	}

out:
	release_firmware(fw);
	return ret;
}

static int spec_irq_request(struct fmc_device *fmc, irq_handler_t handler,
			    char *name, int flags)
{
	return 0;
}

static void spec_irq_ack(struct fmc_device *fmc)
{
	return;
}

static int spec_irq_free(struct fmc_device *fmc)
{
	return 0;
}

static int spec_gpio_config(struct fmc_device *fmc, struct fmc_gpio *gpio,
			    int ngpio)
{
	return 0;
}

/* The engines for this live in spec-i2c.c, we only shape arguments */
static int spec_read_ee(struct fmc_device *fmc, int pos, void *data, int len)
{
	return spec_eeprom_read(fmc, pos, data, len);
}

static int spec_write_ee(struct fmc_device *fmc, int pos,
			 const void *data, int len)
{
	return spec_eeprom_write(fmc, pos, data, len);
}

static struct fmc_operations spec_fmc_operations = {
	/* no readl/writel because we have the base pointer */
	.validate =		spec_validate,
	.reprogram_raw =        spec_reprogram_raw,
	.reprogram =		spec_reprogram,
	.irq_request =		spec_irq_request,
	.irq_ack =		spec_irq_ack,
	.irq_free =		spec_irq_free,
	.gpio_config =		spec_gpio_config,
	.read_ee =		spec_read_ee,
	.write_ee =		spec_write_ee,
};

int spec_fmc_create(struct spec_dev *spec, struct fmc_gateware *gw)
{
	struct fmc_device *fmc;
	struct pci_dev *pdev;
	int ret;

	fmc = kzalloc(sizeof(*fmc), GFP_KERNEL);
	if (!fmc)
		return -ENOMEM;

	fmc->version = FMC_VERSION;
	fmc->owner = THIS_MODULE;
	fmc->carrier_name = "SPEC";
	fmc->carrier_data = spec;

	/* 1M window at offset 0 */
	fmc->fpga_base = spec->remap[0];
	fmc->memlen = 1 << 20;

	fmc->op = &spec_fmc_operations;
	fmc->hwdev = &spec->pdev->dev; /* for messages */
	spec->fmc = fmc;

	/* We have one slot only, and the i2c address is mandated */
	fmc->slot_id = 0;
	fmc->eeprom_addr = SPEC_I2C_EEPROM_ADDR;

	/* The device id is needed to build mezzanine unique names */
	pdev = spec->pdev;
	fmc->device_id = (pdev->bus->number << 8) | pdev->devfn;

	/* Scan the SDB information from the FPGA memory space */
	ret = spec_sdb_load(fmc);
	if (ret)
		goto out_free;

	ret = spec_i2c_init(fmc);
	if (ret)
		goto out_free;
	ret = spec_irq_init(spec);
	if (ret)
		goto out_free;
	ret = fmc_device_register_gw(fmc, gw);
	if (ret)
		goto out_irq;
	spec_gpio_init(fmc); /* May fail, we don't care */

	if (!(fmc->flags &= FMC_DEVICE_HAS_CUSTOM)) {
		if(!strcmp(fmc->id.product_name,DIO_FMC_ID_NAME)) {
			ret = spec_gw_program(fmc,1);
		} else {
			ret = spec_gw_program(fmc,0);
		}

		if (ret)
			goto out_irq;
	}

	ret = spec_sdb_fpga_dev_register_all(fmc);
	if(ret)
		goto out_irq;

	spec->flags |= SPEC_FLAG_FMC_REGISTERED;
	return ret;

out_irq:
	spec_irq_exit(spec);
out_free:
	spec->fmc = NULL;
	kfree(fmc);
	return ret;
}

void spec_fmc_destroy(struct spec_dev *spec)
{
	if (!(spec->flags & SPEC_FLAG_FMC_REGISTERED))
		return;
	/* undo the things in the reverse order, but pin the device first */
	spec_sdb_fpga_dev_release_all(spec->fmc);
	get_device(&spec->fmc->dev);
	spec_gpio_exit(spec->fmc);
	fmc_device_unregister(spec->fmc);
	spec_irq_exit(spec);
	spec_i2c_exit(spec->fmc);
	put_device(&spec->fmc->dev);
	/*
	 * Do not do `kfree(spec->fmc);` because it is already done by
	 * the fmc-bus
	 */
	spec->fmc = NULL;
	spec->flags &= ~SPEC_FLAG_FMC_REGISTERED;
}
