/*
 * Copyright (C) 2019 Seven Solutions (sevensols.com)
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 */

#include <linux/firmware.h>
#include <linux/types.h>
#include <linux/platform_device.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/delay.h>
#include <linux/workqueue.h>
#include <linux/fmc.h>
#include <linux/fmc-sdb.h>

#include "spec.h"

/*
 * SDB information for several FPGA IP blocks.
 */
// SDB MAGIC
#define SDB_MAGIC 0x5344422d
// CERN devices
// Vendor
#define SDB_CERN_VENDOR		0x0000ce42
// Products
#define SDB_NIC_PID			0x00000012
#define SDB_NIC_NAME		"NIC"
#define SDB_EP_PID			0x650c2d4f
#define SDB_EP_NAME			"EP"
#define SDB_TXTSU_PID		0x00000014
#define SDB_TXTSU_NAME		"TxTSU"
#define SDB_PPSG_PID		0xde0d8ced
#define SDB_PPSG_NAME		"PPSG"
#define SDB_VIC_PID			0x00000013
#define SDB_VIC_NAME		"VIC"
#define SDB_GPIO_PID		0x441c5143
#define SDB_GPIO_NAME		"GPIO"
#define SDB_RAM_PID			0x66cfeb52
#define SDB_RAM_NAME		"RAM"
#define SDB_SYSCON_PID		0xff07fc47
#define SDB_SYSCON_NAME		"SYSCON"
// Seven Solutions devices
// Vendor
#define SDB_7SOLS_VENDOR	0x000075cb
// Products
#define SDB_DIO_PID			0x00000001
#define SDB_DIO_NAME		"DIO"

// Designs entry points, bitstreams and firmware files
#define SPEC_GOLDEN_GW_NAME "fmc/spec-init.bin"
#define SDB_SPEC_GOLDEN_ENTRY 0x100 
#define DIO_GW_NAME "fmc/wr_nic_dio.bin"
#define SDB_NIC_DIO_ENTRY 0x70000
#define NIC_GW_NAME "fmc/wr_nic.bin"
#define SDB_NIC_ENTRY 0x70000
#define NIC_LM32_FW_NAME "fmc/wr_nic_lm32_sw.bin"

// Misc
#define RAM_MAGIC 0x98000000
#define SYSCON_RESET_ON 0x1deadbee
#define SYSCON_RESET_OFF 0x0deadbee
/**
 * @brief FPGA IP block structure.
 *
 * This contains basic information in regards of a single FPGA
 * IP block.
 */
struct fpga_ip {
	char *name; /**< Name of the FPGA block*/
	uint64_t vid; /**< Vendor identifier for the FPGA block */
	uint32_t pid; /**< Product identifier for the FPGA block */
	unsigned int irq; /**< FPGA block interrupt if any */
	int has_irq; /**< FPGA block has interrupt? **/
};

// 10 FPGA cores is the maximum value for a single FPGA device
#define FPGA_DEV_MAX_CORES 10
/**
 * @brief FPGA device structure.
 *
 * This contains information regarding FPGA blocks that are
 * required by a specific FPGA device.
 */
struct fpga_dev {
	char *name; /**< Name of the FPGA device */
	struct fpga_ip cores[FPGA_DEV_MAX_CORES]; /**< List of cores required by the FPGA device */
	unsigned int n_cores; /**< Number of elements in the cores list **/

	struct platform_device dev; /**< Linux platform device for the FPGA device **/
	void *priv; /**< Private data for FPGA dev */
	int dev_registered; /**< Linux platform device has been registered into system **/
};

/**
 * Release function for the FPGA platform device.
 *
 *@param dev Linux device for FPGA block
 *
 *@warning It does not require any specific application by now.
 */
static void fpga_dev_release(struct device *dev){}

// Macros for creating a static fpga_dev array
#define FPGA_DEVS_CREATE_BEGIN_DEV(id,name_dev) \
	[id] = {\
		.name = name_dev, \
		.cores = {
#define FPGA_DEVS_CREATE_END_DEV(n_elements,name_pdev,release_pdev) \
	}, \
	.n_cores = n_elements, \
	.dev = { \
		.name = name_pdev, \
		.id = 0, \
		.dev.release = release_pdev \
		} \
	},
#define FPGA_DEV_CREATE_FULL(id,name_dev,vendor_id,product_id,fpga_irq,fpga_has_irq) \
	[id] = {.name = name_dev, \
	.vid = vendor_id, \
	.pid = product_id,\
	.irq = fpga_irq, \
	.has_irq = fpga_has_irq},
	
#define FPGA_DEV_CREATE_SFULL(id,name_dev,vendor_id,product_id,fpga_irq) \
	[id] = {.name = name_dev, \
	.vid = vendor_id, \
	.pid = product_id,\
	.irq = fpga_irq, \
	.has_irq = 1},

#define FPGA_DEV_CREATE(id,name_dev,vendor_id,product_id) \
	FPGA_DEV_CREATE_FULL(id,name_dev,vendor_id,product_id,0,0)

// VIC FPGA device
#define FPGA_DEVS_VIC_DEV 0
#define FPGA_DEVS_VIC_DEV_NAME "spec-vic"
#define FPGA_DEVS_VIC_DEV_VIC 0
#define FPGA_DEVS_VIC_DEV_VIC_IRQ 9
#define FPGA_DEVS_VIC_DEV_NUM 1
#define FPGA_DEVS_VIC_DEV_PDEV_NAME "htvic-spec"
#define FPGA_DEVS_VIC_DEV_PDEV_RELEASE_F fpga_dev_release
// NIC FPGA device
#define FPGA_DEVS_NIC_DEV 1
#define FPGA_DEVS_NIC_DEV_NAME "spec-nic"
#define FPGA_DEVS_NIC_DEV_NIC 0
#define FPGA_DEVS_NIC_DEV_NIC_IRQ 1
#define FPGA_DEVS_NIC_DEV_EP 1
#define FPGA_DEVS_NIC_DEV_TXTSU 2
#define FPGA_DEVS_NIC_DEV_TXTSU_IRQ 0
#define FPGA_DEVS_NIC_DEV_PPSG 3
#define FPGA_DEVS_NIC_DEV_NUM 4
#define FPGA_DEVS_NIC_DEV_PDEV_NAME "spec-nic"
#define FPGA_DEVS_NIC_DEV_PDEV_RELEASE_F fpga_dev_release
// DIO FPGA device
#define FPGA_DEVS_DIO_DEV 2
#define FPGA_DEVS_DIO_DEV_NAME "spec-fmc-dio"
#define FPGA_DEVS_DIO_DEV_DIO 0
#define FPGA_DEVS_DIO_DEV_DIO_IRQ 3
#define FPGA_DEVS_DIO_DEV_GPIO 1
#define FPGA_DEVS_DIO_DEV_PPSG 2
#define FPGA_DEVS_DIO_DEV_NUM 3
#define FPGA_DEVS_DIO_DEV_PDEV_NAME "fmc-dio-spec"
#define FPGA_DEVS_DIO_DEV_PDEV_RELEASE_F fpga_dev_release

// Global structure for FPGA devices
static struct fpga_dev fpga_devs[] =
{
		/* VIC */
		FPGA_DEVS_CREATE_BEGIN_DEV(FPGA_DEVS_VIC_DEV,FPGA_DEVS_VIC_DEV_NAME)
		FPGA_DEV_CREATE_SFULL(FPGA_DEVS_VIC_DEV_VIC,SDB_VIC_NAME,SDB_CERN_VENDOR,SDB_VIC_PID,FPGA_DEVS_VIC_DEV_VIC_IRQ)
		FPGA_DEVS_CREATE_END_DEV(FPGA_DEVS_VIC_DEV_NUM,FPGA_DEVS_VIC_DEV_PDEV_NAME,FPGA_DEVS_VIC_DEV_PDEV_RELEASE_F)
		/* NIC */
		FPGA_DEVS_CREATE_BEGIN_DEV(FPGA_DEVS_NIC_DEV,FPGA_DEVS_NIC_DEV_NAME)
		FPGA_DEV_CREATE_SFULL(FPGA_DEVS_NIC_DEV_NIC,SDB_NIC_NAME,SDB_CERN_VENDOR,SDB_NIC_PID,FPGA_DEVS_NIC_DEV_NIC_IRQ)
		FPGA_DEV_CREATE(FPGA_DEVS_NIC_DEV_EP,SDB_EP_NAME,SDB_CERN_VENDOR,SDB_EP_PID)
		FPGA_DEV_CREATE_SFULL(FPGA_DEVS_NIC_DEV_TXTSU,SDB_TXTSU_NAME,SDB_CERN_VENDOR,SDB_TXTSU_PID,FPGA_DEVS_NIC_DEV_TXTSU_IRQ)
		FPGA_DEV_CREATE(FPGA_DEVS_NIC_DEV_PPSG,SDB_PPSG_NAME,SDB_CERN_VENDOR,SDB_PPSG_PID)
		FPGA_DEVS_CREATE_END_DEV(FPGA_DEVS_NIC_DEV_NUM,FPGA_DEVS_NIC_DEV_PDEV_NAME,FPGA_DEVS_NIC_DEV_PDEV_RELEASE_F)
		/* DIO */
		FPGA_DEVS_CREATE_BEGIN_DEV(FPGA_DEVS_DIO_DEV,FPGA_DEVS_DIO_DEV_NAME)
		FPGA_DEV_CREATE_SFULL(FPGA_DEVS_DIO_DEV_DIO,SDB_DIO_NAME,SDB_7SOLS_VENDOR,SDB_DIO_PID,FPGA_DEVS_DIO_DEV_DIO_IRQ)
		FPGA_DEV_CREATE(FPGA_DEVS_DIO_DEV_GPIO,SDB_GPIO_NAME,SDB_CERN_VENDOR,SDB_GPIO_PID)
		FPGA_DEV_CREATE(FPGA_DEVS_DIO_DEV_PPSG,SDB_PPSG_NAME,SDB_CERN_VENDOR,SDB_PPSG_PID)
		FPGA_DEVS_CREATE_END_DEV(FPGA_DEVS_DIO_DEV_NUM,FPGA_DEVS_DIO_DEV_PDEV_NAME,FPGA_DEVS_DIO_DEV_PDEV_RELEASE_F)

};

/**
 * Register a Linux platform device for a specific FPGA device.
 *
 * This function reads information of the FPGA device, creates resources for each specific FPGA cores
 * and registers the Linux platform device.
 *
 * @param fmc FMC device
 * @param fdev FPGA device
 * @param priv Private data for device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int spec_sdb_fpga_dev_register(struct fmc_device *fmc, struct fpga_dev *fdev)
{
	signed long start;
	unsigned long size;
	struct resource *res = NULL;
	struct spec_dev *spec = fmc->carrier_data;
	int i, j, r;
	unsigned int n_irqs = 0;
	unsigned int n_res = 0;

	if(fdev->n_cores <= 0 || fdev->n_cores >= FPGA_DEV_MAX_CORES)
		return -EINVAL;

	// Count the number of interrupts for the FPGA blocks
	for(i = 0 ; i < fdev->n_cores ; i++)
		if(fdev->cores[i].has_irq)
			n_irqs++;

	n_res = fdev->n_cores+n_irqs;

	res = kzalloc(sizeof(*res)*n_res, GFP_KERNEL);
	if(!res)
		return -ENOMEM;

	for(i = 0, j = 0 ; i < fdev->n_cores ; i++) {
		start = fmc_find_sdb_device(fmc->sdb,
				fdev->cores[i].vid,
				fdev->cores[i].pid,&size);
		if(start < 0) {
			kfree(res);
			return -EINVAL;
		}

		res[j].name = fdev->cores[i].name;
		res[j].flags = IORESOURCE_MEM;
		res[j].start = spec->area[0]->start+start;
		res[j].end = res[j].start+size-1;
		res[j].parent = spec->area[0];

		j++;

		if(fdev->cores[i].has_irq) {
			res[j].name = fdev->cores[i].name;
			res[j].flags = IORESOURCE_IRQ | IORESOURCE_IRQ_HIGHEDGE;
			res[j].start = fdev->cores[i].irq;
			res[j].end = fdev->cores[i].irq;

			j++;
		}

	}

	fdev->dev.resource = res;
	fdev->dev.num_resources = n_res;

	platform_set_drvdata(&fdev->dev,fdev->priv);

	r = platform_device_register(&fdev->dev);
	if(r) {
		fdev->dev_registered = 0;
		kfree(res);
	}
	else {
		fdev->dev_registered = 1;
	}

	return r;
}

//static struct irq_domain * irq_find_irqdomain(uint32_t id)
static struct irq_domain * irq_find_irqdomain(struct device *dev)
{
	struct irq_fwspec fwspec;

	memset(&fwspec,0,sizeof(fwspec));

	//fwspec.param_count = 1;
	//fwspec.param[0] = id;
	
	fwspec.param_count = 2;
	fwspec.param[0] = (uint32_t) (((unsigned long) dev) >> 32);
	fwspec.param[1] = (uint32_t) (((unsigned long) dev) & 0xFFFFFFFF);

	return irq_find_matching_fwspec(&fwspec,DOMAIN_BUS_ANY);
}

struct work_struct fpga_cores_register_wq;
#define FPGA_CORES_REG_WQ_SLEEP_MS 1000
void spec_sdb_fpga_dev_register_wq(struct work_struct *work)
{
	struct device *dev = &fpga_devs[FPGA_DEVS_VIC_DEV].dev.dev;
	struct irq_domain *d;
	struct fmc_device *fmc = fpga_devs[FPGA_DEVS_DIO_DEV].priv;
	int i,j,r;

	// Wait until VIC driver has been installed
	while(dev->driver == NULL) {
		msleep(FPGA_CORES_REG_WQ_SLEEP_MS);
	}

	// Map IRQs for the rest of the modules
	d = irq_find_irqdomain(dev); 
	for(i = 1 ; i < ARRAY_SIZE(fpga_devs) ; i++) {
		for(j = 0 ; j < fpga_devs[i].n_cores ; j++) {
			if(fpga_devs[i].cores[j].has_irq) {
				fpga_devs[i].cores[j].irq = irq_find_mapping(d,
						fpga_devs[i].cores[j].irq);
			}
		}
	}

	// Register the rest of modules
	for(i = 1 ; i < ARRAY_SIZE(fpga_devs) ; i++) {
		r = spec_sdb_fpga_dev_register(fmc,&fpga_devs[i]);
		if(r != 0)
			dev_err(&fmc->dev,"%s: %s could not be registered! \n",
					__func__,fpga_devs[i].name);
	}
}


/**
 * Register Linux platform devices for the all FPGA resources.
 *
 * @param fmc FMC device
 *
 * @return 0 if success and a negative error code otherwise
 */
int spec_sdb_fpga_dev_register_all(struct fmc_device *fmc)
{
	int i;
	int r;
	struct spec_dev *spec = fmc->carrier_data;
	struct irq_domain *domain = spec->gpio_domain;

	// Init workqueue for the delayed registration
	INIT_WORK(&fpga_cores_register_wq, spec_sdb_fpga_dev_register_wq);

	// Set private information
	for(i = 0 ; i < ARRAY_SIZE(fpga_devs) ; i++)
		fpga_devs[i].priv = (i == FPGA_DEVS_DIO_DEV) ?
				(void *) fmc : fmc->fpga_base;

	// Find Linux IRQ for VIC
	fpga_devs[FPGA_DEVS_VIC_DEV].cores[FPGA_DEVS_VIC_DEV_VIC].irq =
			irq_find_mapping(domain, FPGA_DEVS_VIC_DEV_VIC_IRQ);
	// Register VIC
	r = spec_sdb_fpga_dev_register(fmc,&fpga_devs[FPGA_DEVS_VIC_DEV]);
	if(r != 0)
		dev_err(&fmc->dev,"%s: %s could not be registered! \n",
				__func__,fpga_devs[0].name);

	// Schedule worqueue
	schedule_work(&fpga_cores_register_wq);

	return 0;
}

/**
 * Release Linux platform devices for the all FPGA resources.
 *
 * @param fmc FMC device
 */
void spec_sdb_fpga_dev_release_all(struct fmc_device *fmc)
{
	int i;
	struct fpga_dev *dev;

	for(i = 0 ; i < ARRAY_SIZE(fpga_devs) ; i++) {
		dev = &fpga_devs[i];
		if(dev->dev_registered) {
			platform_device_unregister(&dev->dev);
			kfree(dev->dev.resource);
		}
	}

	return;
}

/**
 * Initialize SDB structure for the FMC device.
 *
 * @param fmc FMC device
 * 
 * @return 0 if success and a negative error code otherwise
 */
int spec_sdb_load(struct fmc_device *fmc)
{
	struct spec_dev *spec = fmc->carrier_data;
	unsigned long sdb_entry = spec->sdb_entry;
	int ret;

	/* poor man's SDB */
	if (fmc_readl(fmc, sdb_entry) != SDB_MAGIC) {
		dev_err(&spec->pdev->dev, "Can't find SDB magic\n");
		return -ENODEV;
	}

	ret = fmc_scan_sdb_tree(fmc, sdb_entry);
	if (ret < 0)
		return -ENODEV;

	spec->syscon_addr = fmc_find_sdb_device(fmc->sdb,
			SDB_CERN_VENDOR, SDB_SYSCON_PID,
			&spec->syscon_size);
	if(spec->syscon_addr < 0)
		return -ENODEV;

	/* It is not a golden gateware, mark it. */
	if(sdb_entry != SDB_SPEC_GOLDEN_ENTRY)
		fmc->flags |= FMC_DEVICE_HAS_CUSTOM;
	return 0;
}

/**
 * Write LM32 RAM.
 *
 * @param fmc FMC device
 * @param name LM32 firmware name
 * @param ram Physical address for LM32 RAM
 * @param ramsize Physical size for LM32 RAM
 * 
 * @return 0 if success and a negative error code otherwise
 */
static int spec_lm32_fw_load(struct fmc_device *fmc, char *name,
			unsigned long ram, unsigned long ramsize)
{
	const struct firmware *fw;
	struct device *dev = fmc->hwdev;
	int ret, count;
	const uint32_t *p;

	ret = request_firmware(&fw, name, dev);
	if (ret < 0) {
		dev_err(dev, "request firmware \"%s\": error %i\n",
			name, ret);
		return ret;
	}
	if (fw->size > ramsize) {
		dev_err(dev, "firmware \"%s\" longer than ram (0x%lx)\n",
			name, ramsize);
		ret = -ENOMEM;
		goto out;
	}
	for (count = 0, p = (void *)fw->data; count < fw->size; count += 4, p++)
		fmc_writel(fmc, __cpu_to_be32(*p), ram + count);
out:
	release_firmware(fw);
	return ret;
}

/**
 * Program FPGA Gateware including NIC with or without DIO support.
 *
 * @param fmc FMC device
 * @param has_dio Include DIO support (=1) or only NIC.
 * 
 * @return 0 if success and a negative error code otherwise
 */
int spec_gw_program(struct fmc_device *fmc, int has_dio)
{
	int ret = 0;
	struct spec_dev *spec = fmc->carrier_data;
	struct device *dev = &spec->pdev->dev;
	signed long ram, syscon;
	unsigned long ramsize;
	char *filename = NIC_LM32_FW_NAME;
	unsigned long j;

	/* First, free SDB tree */
	fmc_free_sdb_tree(fmc);

	/* Program FPGA bitstream */
	if(has_dio) {
		spec->sdb_entry = SDB_NIC_DIO_ENTRY;
		ret = spec_load_fpga_file(spec,DIO_GW_NAME);
	} else {
		spec->sdb_entry = SDB_NIC_ENTRY;
		ret = spec_load_fpga_file(spec,NIC_GW_NAME);
	}

	if(ret)
		goto out;

	/* Load SDB information */
	ret = spec_sdb_load(fmc);
	if(ret)
		goto out;

	/* Write LM32 firmware for WR-NIC design */
	ram = fmc_find_sdb_device(fmc->sdb, SDB_CERN_VENDOR, SDB_RAM_PID, &ramsize);
	syscon = spec->syscon_addr;

	j = jiffies + HZ/2;
	fmc_writel(fmc, SYSCON_RESET_ON, syscon);
	while ( !(fmc_readl(fmc, syscon) & (1 << 28)) )
		if (time_after(jiffies, j))
			break;
	if (time_after(jiffies, j)) {
		dev_err(dev, "can't reset LM32\n");
		fmc_writel(fmc, SYSCON_RESET_OFF, syscon);
		goto out;
	}
	ret = spec_lm32_fw_load(fmc, filename, ram, ramsize);
	fmc_writel(fmc, SYSCON_RESET_OFF, syscon);
	if (ret)
		goto out; /* message already reported */
	if (fmc_readl(fmc, ram) != RAM_MAGIC)
		dev_warn(dev, "possible failure in wrc load\n");
	else
		dev_info(dev, "WRC program reloaded from \"%s\"\n",
			 filename);

	return 0;

out:
	return ret;
}
