/* A simple console for accessing the SPEC virtual UART (i.e. for communicating with the WR Core shell
   from a Linux terminal. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <errno.h>

#include "speclib.h"

static char git_version[] = "version: " GIT_VERSION;

static void print_version(char *pname)
{
	printf("%s %s\n", pname, git_version);
	printf("%s\n", libspec_version_s);
}

static void *card;

void term_main(int keep_term, int cmd_mode, char * cmd)
{
	struct termios oldkey, newkey;
	//above is place for old and new port settings for keyboard teletype
	int need_exit = 0;
	char c;
	int prompt_det = 0;
	char tmp[20];
	int tmp_idx;
	int i;

	if (cmd_mode == 0)
		fprintf(stderr, "[press C-a to exit]\n");

	if(!keep_term) {
		tcgetattr(STDIN_FILENO,&oldkey);
		newkey.c_cflag = B115200 | CS8 | CLOCAL | CREAD;
		newkey.c_iflag = IGNPAR;
		newkey.c_oflag = 0;
		newkey.c_lflag = 0;
		newkey.c_cc[VMIN]=1;
		newkey.c_cc[VTIME]=0;
		tcflush(STDIN_FILENO, TCIFLUSH);
		tcsetattr(STDIN_FILENO,TCSANOW,&newkey);
	}
	
	if (cmd_mode == 1) {
		char endc;
		
		vuart_flush(card);
		
		spec_vuart_tx(card,cmd,strlen(cmd));
		
		do {
			spec_vuart_rx(card, &endc, 1);
		} while(endc != 13);
	}
	
	while(!need_exit) {
		fd_set fds;
		int ret;
		char rx;
		struct timeval tv = {0, 10000};

		FD_ZERO(&fds);
		FD_SET(STDIN_FILENO, &fds);

		ret = select(STDIN_FILENO+1, &fds, NULL, NULL, &tv);
		if(ret == -1) {
			perror("select");
		} else if (ret > 0) {
			if(FD_ISSET(STDIN_FILENO, &fds)) {
				need_exit = vuart_check_end(STDIN_FILENO, 1, &c);
				
				if(need_exit == 0 && cmd_mode == 0) {
					spec_vuart_tx(card, &c, 1);
				}
				else {
					if(need_exit == 0 && cmd_mode == 1 && c == 27) {
						spec_vuart_tx(card, &c, 1);
					}
				}
			}
		}

		prompt_det = 0;
		tmp_idx = 0;
		while((spec_vuart_rx(card, &rx, 1)) == 1) {
			
			#if VUART_RX_USLEEP > 0
				usleep(VUART_RX_USLEEP);
			#endif
			
			if(cmd_mode == 1) {
				vuart_detect_prompt(rx,&prompt_det,VUART_PROMPT_DET_MODE_FULL);

				switch(prompt_det)
				{
				case VUART_PROMT_DETECTED:
					need_exit=-1;
					break;
				case VUART_W_DETECTED:
					tmp[tmp_idx] = rx;
					tmp_idx++;
					break;
				case VUART_R_DETECTED:
					tmp[tmp_idx] = rx;
					tmp_idx++;
					break;
				case VUART_C_DETECTED:
					tmp[tmp_idx] = rx;
					tmp_idx++;
					break;
				default:
					for(i = 0 ; i < tmp_idx ; i++)
						fprintf(stderr,"%c",tmp[tmp_idx]);
					tmp_idx = 0;
					fprintf(stderr,"%c",rx);
					break;
				}
			} else {
				fprintf(stderr,"%c", rx);
			}
			
		}

	}

	if(!keep_term)
		tcsetattr(STDIN_FILENO,TCSANOW,&oldkey);
}

int main(int argc, char **argv)
{
	int bus = -1, dev_fn = -1, c;
	uint32_t vuart_base = 0x20500;
	int keep_term = 0;
	int cmd_len;
	int cmd_mode=0;
	char cmd[VUART_SIZE_CMD];

	while ((c = getopt (argc, argv, "b:d:u:c:kV")) != -1)
	{
		switch(c)
		{
		case 'b':
			sscanf(optarg, "%i", &bus);
			break;
		case 'd':
			sscanf(optarg, "%i", &dev_fn);
			break;
		case 'u':
			sscanf(optarg, "%i", &vuart_base);
			break;
		case 'k':
			keep_term = 1;
			break;
		case 'c':
			cmd_mode=1;
			strcpy(cmd,optarg);
			cmd_len = strlen(cmd);
			cmd[cmd_len] = 13;
			cmd[cmd_len+1] = 0;
			break;
		case 'V':
			print_version(argv[0]);
			exit(0);
		default:
			fprintf(stderr,
				"Use: \"%s [-V] [-b bus] [-d devfn] "
				"[-u VUART base] [-k]\"\n", argv[0]);
			fprintf(stderr,
				"By default, the first available SPEC "
				"is used and the VUART is assumed at 0x%x.\n"
				"-k option keeps the prev terminal config\n",
				vuart_base);
			exit(1);
		}
	}

    card = spec_open(bus, dev_fn);
	if(!card)
	{
		fprintf(stderr, "Can't detect a SPEC card under the given "
			"adress.\nMake sure a SPEC card is present in your PC, "
			"the driver is loaded and you run the program as root.\n");
		exit(1);
	}

	spec_vuart_init(card, vuart_base);
	term_main(keep_term,cmd_mode,cmd);
	spec_close(card);

	return 0;
}
